import ast
import os
from collections import defaultdict

from pydantic import BaseModel, PrivateAttr

from pycodedoc.llm import Llm
from pycodedoc.parser import Parser
from pycodedoc.prompts import (
    PROMPTS,
    get_classes_prompts,
    get_functions_prompts,
    get_modules_deps_prompts,
    get_modules_prompts,
    get_project_prompt,
)
from pycodedoc.utils import set_logger

logger = set_logger()


class Descriptions(BaseModel):
    entities: dict = defaultdict(dict)
    functions: dict = defaultdict(dict)
    classes: dict = defaultdict(dict)
    modules: dict = defaultdict(dict)
    modules_deps: dict = defaultdict(dict)
    project: str = ""


class DocGen(BaseModel, extra="forbid"):
    """
    The DocGen class is responsible for generating documentation for a Python project.
    It uses the OpenAI model to generate descriptions for functions, classes, and modules.
    It also creates execution graphs if Graphviz is installed.

    Attributes:
        base_dir (str): The directory of the Python project you want to document.
        use_structure (bool): Use the structure of the code to generate the documentation. Default is False.
        no_relations (bool): Add description of the relationship between modules. Default is True.
        no_classes (bool): Create documentation for classes. Default is True.
        create_graphs (bool): Create execution graphs of the code. Default is True.
        prompts (dict): The prompts for the OpenAI model.
        output_dir (str): The path of the output directory. Default is "./docs".
        model (str): The OpenAI model to use for generating the documentation. Default is "gpt-3.5-turbo-0125".
        llm (Llm): The language model.
        parser (Parser): The parser for the Python code.
        _descriptions (Descriptions): The descriptions generated by the OpenAI model.
    """

    base_dir: str
    use_structure: bool = False
    no_relations: bool = False
    no_classes: bool = False
    create_graphs: bool = True
    prompts: dict = PROMPTS
    output_dir: str = "./docs"
    model: str = "gpt-3.5-turbo-0125"
    llm: Llm = Llm()
    parser: Parser = None
    _descriptions: Descriptions = PrivateAttr(Descriptions())

    def model_post_init(self, __context):
        self.parser = Parser(base_dir=self.base_dir)

    def generate_documentation(self):
        """
        Generates the documentation for the Python project.

        This method generates descriptions for functions, classes, and modules based on the attributes set in the DocGen instance.
        It also generates descriptions of the relationships between modules and an overview of the project.
        Finally, it writes all the generated descriptions to a markdown file.

        The generation of descriptions for functions and classes can be toggled on or off using the `use_structure` and `no_classes` attributes respectively.
        The generation of descriptions for the relationships between modules can be toggled on or off using the `no_relations` attribute.
        """
        if self.use_structure:
            logger.info("GENERATING FUNCTIONS DESCRIPTIONS")
            self.generate_functions_desc()
        if not self.no_classes:
            logger.info("GENERATING CLASSES DESCRIPTIONS")
            self.generate_classes_desc()
        logger.info("GENERATING MODULES DESCRIPTIONS")
        self.generate_modules_desc()
        if not self.no_relations:
            logger.info("GENERATING MODULES RELATIONS DESCRIPTIONS")
            self.generate_modules_deps_desc()
        logger.info("GENERATING PROJECT OVERVIEW")
        self.generate_project_desc()
        self.write_markdown()

    def generate_descriptions(self, attr: str, module_path: str = None):
        """
        Generate descriptions for the specified attribute.

        Args:
            attr (str): The attribute for which to generate descriptions. Can be one of the following: functions, classes, modules, modules_relations, project.
            module_path (str, optional): The path of the module for which to generate descriptions. Defaults to None.

        Returns:
            Descriptions: The descriptions generated by the OpenAI model.

        Raises:
            ValueError: If the attribute is not recognized.
        """
        if attr == "functions":
            self.generate_functions_desc(module_path)
        elif attr == "classes":
            self.generate_classes_desc(module_path)
        elif attr == "modules":
            self.generate_modules_desc(module_path)
        elif attr == "modules_relations":
            self.generate_modules_deps_desc(module_path)
        elif attr == "project":
            self.generate_project_desc()
        else:
            raise ValueError(
                f"Attribute {attr} not recognized. Please use one of the following: functions, classes, modules, modules_relations, project."
            )
        return self.get_descriptions(attr)

    def get_descriptions(self, attr: str = None):
        if attr is None:
            return self._descriptions
        else:
            return getattr(self._descriptions, attr)

    def generate_functions_desc(self, module_path: str = None):
        functions_code = self.parser.get_functions(module_path, attr="code")
        prompts = get_functions_prompts(functions_code, **self.prompts["functions"])
        responses = self.llm.run_batch_completions(
            **prompts, timeout=10, stream=True, model=self.model
        )
        functions = self.parser.get_functions(module_path)
        for function, response in zip(functions, responses):
            self._descriptions.entities[function.path][function.uname] = response[
                "content"
            ]
            self._descriptions.functions[function.path][function.uname] = response[
                "content"
            ]

    def generate_classes_desc(self, module_path: str = None):
        classes = self.parser.get_classes(module_path)
        classes_code = self.get_classes_code(classes)
        prompts = get_classes_prompts(classes_code, **self.prompts["classes"])
        responses = self.llm.run_batch_completions(
            **prompts, timeout=10, stream=True, model=self.model
        )
        for class_, response in zip(classes, responses):
            self._descriptions.entities[class_.path][class_.name] = response["content"]
            self._descriptions.classes[class_.path][class_.name] = response["content"]

    def get_classes_code(self, classes):
        classes_code = []
        for class_ in classes:
            if self.use_structure:
                descriptions = self._descriptions.entities[class_.path]
                code = self.parser.get_code_structure(class_, descriptions=descriptions)
            else:
                code = ast.unparse(class_.node)
            classes_code.append(code)
        return classes_code

    def generate_modules_desc(self, module_path: str = None):
        modules = self.parser.get_modules(module_path)
        modules_code = self.get_modules_code(modules)
        prompts = get_modules_prompts(modules_code, **self.prompts["modules"])
        responses = self.llm.run_batch_completions(
            **prompts, timeout=10, stream=True, model=self.model
        )
        for module, response in zip(modules, responses):
            self._descriptions.modules[module.path] = response["content"]

    def get_modules_code(self, modules):
        modules_code = []
        for module in modules:
            if self.use_structure:
                descriptions = self._descriptions.entities[module.path]
                code = self.parser.get_code_structure(module, descriptions=descriptions)
            else:
                code = ast.unparse(module.node)
            if self.create_graphs:
                graphviz_installed = self.parser.write_graphs(module, self.output_dir)
                if graphviz_installed is False:
                    self.create_graphs = False
            modules_code.append(code)
        return modules_code

    def generate_modules_deps_desc(self, module_path: str = None):
        modules = self.parser.get_modules(module_path)
        modules_paths, modules_code, deps_code, execution_graphs = [], [], [], []
        for module in modules:
            deps = self.parser.get_module_deps(module.path)
            if any(deps):
                module_code, dep_code, execution_graph = self.parser.get_deps_code(
                    module, deps, self.output_dir, self.create_graphs
                )
                if execution_graph != "":
                    modules_code.append(module_code)
                    deps_code.append(dep_code)
                    execution_graphs.append(execution_graph)
                    modules_paths.append(module.path)
                else:
                    self._descriptions.modules_deps[module.path] = None
            else:
                self._descriptions.modules_deps[module.path] = None
        prompts = get_modules_deps_prompts(
            modules_code, deps_code, execution_graphs, **self.prompts["modules_deps"]
        )
        responses = self.llm.run_batch_completions(
            **prompts, timeout=10, stream=True, model=self.model
        )
        for module_path, response in zip(modules_paths, responses):
            self._descriptions.modules_deps[module_path] = response["content"]

    def generate_project_desc(self):
        modules_docu = self.get_modules_descriptions()
        prompt = get_project_prompt(
            modules_docu, self.parser.get_tree(), **self.prompts["project"]
        )
        response = self.llm.run_completions(
            **prompt, timeout=10, stream=True, model=self.model
        )
        self._descriptions.project = response["content"]

    def get_modules_descriptions(self):
        modules_docu = ""
        if not self.no_relations:
            for module_path in self._descriptions.modules.keys():
                module_desc = self._descriptions.modules[module_path]
                module_deps_desc = self._descriptions.modules_deps[module_path]
                modules_docu += (
                    f"\n\n**Module {module_path}**:\n\nDescription:\n{module_desc}\n"
                )
                if module_deps_desc:
                    modules_docu += (
                        f"\nRelations with other modules:\n{module_deps_desc}\n"
                    )
        else:
            for module_path, module_desc in self._descriptions.modules.items():
                modules_docu += (
                    f"\n\n**Module {module_path}**:\n\nDescription:\n{module_desc}\n"
                )
        return modules_docu

    def get_classes_descriptions(self):
        classes_docu = ""
        for class_path, class_ in self._descriptions.classes.items():
            for class_name, class_desc in class_.items():
                classes_docu += (
                    f"\n**class {class_name} [{class_path}]**:\n\n{class_desc}\n"
                )
        return classes_docu

    def write_markdown(self):
        logger.info("WRITING MARKDOWN DOCUMENTATION")
        md = self.generate_markdown()
        with open(os.path.join(self.output_dir, "project-doc.md"), "w") as f:
            f.write(md)

    def generate_markdown(self):
        md = "# PROJECT OVERVIEW\n\n"
        md += f"{self.get_descriptions('project')}\n\n"

        md += "## PROJECT STRUCTURE\n\n"
        md += f"```\n{self.parser.get_tree()}```\n\n"

        md += "## MODULES"
        md += self.get_modules_descriptions()

        if not self.no_classes:
            md += "\n\n## CLASSES\n"
            md += self.get_classes_descriptions()

        if self.create_graphs:
            title = "\n\n## EXECUTION FLOWS\n\n### MODULES\n"
            graph_md = title
            for module in self.parser.get_modules_paths():
                file_path = (
                    f"{self.output_dir}/graphs/{os.path.splitext(module)[0]}.png"
                )
                if os.path.exists(file_path):
                    rel_path = os.path.relpath(file_path, self.output_dir)
                    graph_md += f"\n![Alt text]({rel_path})\n"
            # only add section if some graphs were added
            if graph_md != title:
                md += graph_md

            if not self.no_relations:
                title = "\n### MODULE RELATIONSHIPS\n"
                graph_md = title
                for module in self.parser.get_modules_paths():
                    file_path = f"{self.output_dir}/graphs/{os.path.splitext(module)[0]}_deps.png"
                    if os.path.exists(file_path):
                        rel_path = os.path.relpath(file_path, self.output_dir)
                        graph_md += f"\n![Alt text]({rel_path})\n"
                # only add section if some graphs were added
                if graph_md != title:
                    md += graph_md

        return md
